/*
 * ttyio.c -- Unix terminal I/O.
 *
 * The functions in this file negotiate with the operating system for
 * keyboard characters, and write characters to the display in a barely 
 * buffered fashion.
 *
 * This is a System V version.  We have use struct termio, but we
 * use other stuff (like FIONREAD, TIOCGWINSZ, and SIGWINCH) if we
 * have it.
 */

/* order is important here... */
#include	<stdio.h>
#include	"ce.h"
#include	<signal.h>
#include	<sys/types.h>
#include	<fcntl.h>
#include	<termio.h>

#define	NOBUF	512			/* Output buffer size.		*/

extern char	*getenv();

static char	obuf[NOBUF];		/* Output buffer.		*/
static int	nobuf;			/* # of chars in output buffer	*/

static struct	termio	ot;
static struct	termio	nt;

static int	ttactive = FALSE;
static int	ttsaved = FALSE;

#ifdef TIOCGWINSZ
static struct 	winsize win;
#endif

/*
 * Resize the screen on receipt of a SIGWINCH, assuming that the screen has
 * actually changed its size.  (Window systems have been known to send 
 * spurious resize notices).
 */
#ifdef SIGWINCH
winresize()
{
	refresh(TRUE, 1);
} 
#endif

#ifndef TCSADRAIN
#define TCSADRAIN 0
#define TCSAFLUSH 1
#define TCSANOW 2
#endif

#define tcgetattr(fd, termiop)	(ioctl((fd), TCGETA, (termiop)))
#define tcsetattr(fd, flag, termiop) \
	(ioctl((fd), \
	       (flag == TCSADRAIN) ? TCSETAW : \
				     ((flag == TCSANOW) ? TCSETA : TCSETAF), \
	       (termiop)))

/*
 * This function gets called once, to set up the terminal channel.  
 * Use 8-bit RAW mode, and set all special characters to 0xFF.
 */
ttraw()
{
	extern short ospeed;

	if (ttactive)
		return;

	if (ttsaved == FALSE) {
		if (tcgetattr(0, &ot) < 0) {
			mlwrite("ttraw can't get initial settings!");
			return(FALSE);
		}
		nt = ot;

		nt.c_cc[VMIN] = 1;
		nt.c_cc[VTIME] = 0;

		nt.c_iflag |= IGNBRK;
		nt.c_iflag &= ~(ICRNL | INLCR | ISTRIP | IXON | IXOFF);
		nt.c_oflag &= ~OPOST;
		nt.c_cflag |= CS8;
		nt.c_cflag &= ~PARENB;
		nt.c_lflag &= ~(ECHO | ICANON | ISIG);

		ttsaved++;
	}
	if (tcsetattr(0, TCSADRAIN, &nt) < 0) {	/* initialized above */
		mlwrite("ttraw can't set terminal!");
		return(FALSE);
	}

	ttactive = TRUE;

	ttsetsize();
#ifdef SIGWINCH
	(void) signal(SIGWINCH, winresize);
#endif
	return(TRUE);
}

ttopen()
{
	if (ttraw() == FALSE) 
		panic("Terminal Initialization Failure!");
}

/*
 * This function gets called just before we go back home to the shell. 
 * Put all of the terminal parameters back.
 */
ttcooked()
{
	ttflush();
	if (ttsaved == 0 || ttactive == 0)
		return(TRUE);
	if (tcsetattr(0, TCSADRAIN, &ot) < 0) {
		mlwrite("ttcooked can't reset terminal!");
		return(FALSE);
	}
	ttactive = FALSE;
	return(TRUE);
}

ttclose()
{
	if (ttcooked() == FALSE) 
		panic("");
}

/*
 * Write character to the display.
 * Characters are buffered up, to make things a little bit more efficient.
 */
ttputc(c)
{
	if (nobuf >= NOBUF)
		ttflush();
	obuf[nobuf++] = c;
}

/*
 * Flush output.
 */
ttflush()
{
	if (nobuf != 0) {
		(void) write(1, obuf, nobuf);
		nobuf = 0;
	}
}

/*
 * Read character from terminal.
 * All 8 bits are returned, so that you can use a multi-national terminal.
 */
ttgetc()
{
	char	buf[1];

	while (read(0, &buf[0], 1) != 1)
		;
	return (buf[0] & 0xFF);
}

/*
 * Set the number of rows and columns on the terminal via the ioctl 
 * provided by 4.3 BSD (and adopted by SunOS, Ultrix, etc.).  If that
 * fails, or is not defined, we try for $LINES and $COLUMNS.  If all
 * fails, we fall back to the termcap entry (though that's usually
 * wrong for xterms).  The number of rows is set to the window size - 1
 * to leave room for the message line, which has its own set of routines.
 */
ttsetsize()
{
#ifdef TIOCGWINSZ
	term.t_nrow = term.t_ncol = -1;
	if (ioctl(0, TIOCGWINSZ, (caddr_t)&win) == 0) {
		term.t_nrow = win.ws_row;
		term.t_ncol = win.ws_col;
	}
#else
	char	*lines, *cols;

	term.t_nrow = term.t_ncol = -1;
	lines = getenv("LINES");
	cols = getenv("COLUMNS");
	if (lines)
		term.t_nrow = atoi(lines);
	if (cols)
		term.t_ncol = atoi(cols);
#endif /* TIOCGWINSZ */

	if (term.t_nrow <= 0)
		term.t_nrow = tgetnum("li");
	if (term.t_ncol <= 0)
		term.t_ncol = tgetnum("co");

	if (term.t_nrow <= 0)
		term.t_nrow = 23;
	else
		term.t_nrow--;
	if (term.t_ncol <= 0)
		term.t_ncol = 80;
}

panic(s)
char *s;
{
	if (ttsaved && ttactive)
		ttcooked();
	xpanic(s);
}

typeahead()
{
#ifdef FIONREAD
	int x;

	return ((ioctl(0, FIONREAD, (caddr_t)&x) < 0) ? 0 : x);
#else
	return 0;	/* if we can't check, there's never any typeahead */
#endif /* FIONREAD */
}

int
cansusp()
{
#ifdef JOBS
	return(1);
#else
	return(0);
#endif /* JOBS */
}
